#include "NeighborsFrame.h"

using namespace std;

NeighborsFrame::NeighborsFrame(const wxString& title, const wxPoint& pos, const wxSize& size) :
    wxFrame(NULL, wxID_ANY, title, pos, size),
    shortestPath(0), waitingForPair(false), startWord(""), endWord(""), wordManager(new WordManager()), matchObject(Match::instance()) {

  // Make sure the dictionary got loaded
  if (matchObject->getDictionary().size() == 0) {
    wxMessageBox(_("There was a problem loading the dictionary file. The program will close."), _("Word Neighbors"), wxICON_ERROR);
    Close(TRUE);
  }

  // Menu Bar
  menuFile = new wxMenu;
  menuFile->Append(ID_NewGame, _("&New Game"));
  menuFile->Append(ID_PathCalculator, _("Show &Path Calculator"));
  menuFile->AppendSeparator();
  menuFile->Append(ID_Quit, _("E&xit"));
  menuHelp = new wxMenu;
  menuHelp->Append(ID_About, _("&About..."));
  menuHelp->AppendSeparator();
  menuHelp->Append(ID_Help, _("&Help"));
  menuBar = new wxMenuBar;
  menuBar->Append(menuFile, _("&File"));
  menuBar->Append(menuHelp, _("&Help"));
  SetMenuBar(menuBar);

  // Tool Bar
  CreateToolBar();
  wxImage::AddHandler(new wxGIFHandler);
  wxImage::AddHandler(new wxPNGHandler);
  wxBitmap playBitmap(_("13555.actionengine_play.gif"), wxBITMAP_TYPE_GIF);
  wxBitmap pathBitmap(_("16060.twowaycompare_co.gif"), wxBITMAP_TYPE_GIF);
  wxBitmap helpBitmap(_("10799.question.gif"), wxBITMAP_TYPE_GIF);
  SetIcon(wxIcon(_("10073.wn_nav.png"), wxBITMAP_TYPE_PNG));
  GetToolBar()->AddTool(ID_NewGame, _("New Game"), playBitmap, _("New Game"));
  GetToolBar()->AddTool(ID_PathCalculator, _("Path Calculator"), pathBitmap, _("Path Calculator"), wxITEM_CHECK);
  GetToolBar()->AddSeparator();
  GetToolBar()->AddTool(ID_Help, _("Help"), helpBitmap, _("Help"));
  GetToolBar()->Realize();

  // Font
  wxFont defaultFont = wxSystemSettings::GetFont(wxSYS_DEFAULT_GUI_FONT);
  defaultFont.SetWeight(wxBOLD);

  panel = new wxPanel(this);
  panelTop = new wxPanel(panel);
  panelBottom = new wxPanel(panel);
  labelStartWordDescriptionTop = new wxStaticText(panelTop, wxID_ANY, _("Starting Word"));
  labelStartWordDescriptionTop->SetFont(defaultFont);
  textBoxStartWord = new wxTextCtrl(panelTop, wxID_ANY, wxEmptyString);
  labelEndWordDescriptionTop = new wxStaticText(panelTop, wxID_ANY, _("Ending Word"));
  labelEndWordDescriptionTop->SetFont(defaultFont);
  textBoxEndWord = new wxTextCtrl(panelTop, wxID_ANY, wxEmptyString);
  buttonFindPath = new wxButton(panelTop, ID_Button, _("Find Path"));
  buttonFindPath->SetToolTip(_("Find Shortest Possible Path"));
  buttonFindPath->SetDefault();
  labelFindPathResultDescription = new wxStaticText(panelTop, wxID_ANY, _("Shortest Path"));
  labelFindPathResultDescription->SetFont(defaultFont);
  labelFindPathResult = new wxStaticText(panelTop, wxID_ANY, wxEmptyString);
  line = new wxStaticLine(panel);
  labelStartWordDescriptionBottom = new wxStaticText(panelBottom, wxID_ANY, _("Starting Word"));
  labelStartWordDescriptionBottom->SetFont(defaultFont);
  labelStartWord = new wxStaticText(panelBottom, wxID_ANY, wxEmptyString);
  labelEndWordDescriptionBottom = new wxStaticText(panelBottom, wxID_ANY, _("Ending Word"));
  labelEndWordDescriptionBottom->SetFont(defaultFont);
  labelEndWord = new wxStaticText(panelBottom, wxID_ANY, wxEmptyString);
  labelWordListDescription = new wxStaticText(panelBottom, wxID_ANY, _("Path"));
  labelWordListDescription->SetFont(defaultFont);
  labelWordList = new wxStaticText(panelBottom, wxID_ANY, wxEmptyString);
  labelWordListStatus = new wxStaticText(panelBottom, wxID_ANY, _("Select Next Word"));
  combo = new wxComboBox(panelBottom, ID_ComboBox);
  combo->SetEditable(false);
  buttonUndo = new wxButton(panelBottom, ID_ButtonUndo, _("Undo"));
  buttonRedo = new wxButton(panelBottom, ID_ButtonRedo, _("Redo"));
  buttonUndo->Disable();
  buttonRedo->Disable();

  // Layout
  sizerAll = new wxBoxSizer(wxVERTICAL);
  sizerTop = new wxBoxSizer(wxVERTICAL);
  sizerBottom = new wxBoxSizer(wxVERTICAL);
  panel->SetSizer(sizerAll, false);
  panelTop->SetSizer(sizerTop, false);
  panelBottom->SetSizer(sizerBottom, false);
  sizerAll->Add(panelTop, 0, wxGROW|wxALL, 2);
  sizerAll->Add(line, 0, wxGROW|wxALL, 2);
  sizerAll->Add(panelBottom, 0, wxGROW|wxALL, 2);
  sizerTop->Add(labelStartWordDescriptionTop, 0, wxGROW|wxALL, 2);
  sizerTop->Add(textBoxStartWord, 0, wxGROW|wxALL, 2);
  sizerTop->Add(labelEndWordDescriptionTop, 0, wxGROW|wxALL, 2);
  sizerTop->Add(textBoxEndWord, 0, wxGROW|wxALL, 2);
  sizerTop->Add(buttonFindPath, 0, wxALL|wxCENTER, 2);
  sizerTop->Add(labelFindPathResultDescription, 0, wxGROW|wxALL, 2);
  sizerTop->Add(labelFindPathResult, 0, wxGROW|wxALL, 2);
  sizerTop->AddSpacer(50);
  sizerWordsHorizontal = new wxBoxSizer(wxHORIZONTAL);
  sizerStartVertical = new wxBoxSizer(wxVERTICAL);
  sizerEndVertical = new wxBoxSizer(wxVERTICAL);
  sizerStartVertical->Add(labelStartWordDescriptionBottom, 0, wxGROW|wxALL, 2);
  sizerStartVertical->Add(labelStartWord, 0, wxGROW|wxALL, 2);
  sizerEndVertical->Add(labelEndWordDescriptionBottom, 0, wxGROW|wxALL, 2);
  sizerEndVertical->Add(labelEndWord, 0, wxGROW|wxALL, 2);
  sizerBottom->Add(sizerWordsHorizontal, wxALL|wxCENTER, 2);
  sizerWordsHorizontal->Add(sizerStartVertical, wxALL|wxCENTER, 2);
  sizerWordsHorizontal->AddSpacer(30);
  sizerWordsHorizontal->Add(sizerEndVertical, wxALL|wxCENTER, 2);
  sizerBottom->AddSpacer(10);
  sizerBottom->Add(labelWordListDescription, 0, wxGROW|wxALL, 2);
  sizerBottom->Add(labelWordList, 0, wxGROW|wxALL, 2);
  sizerBottom->AddSpacer(50);
  sizerBottom->Add(labelWordListStatus, 0, wxGROW|wxALL, 2);
  sizerBottom->Add(combo, 0, wxGROW|wxALL, 2);
  sizerButtonsHorizontal = new wxBoxSizer(wxHORIZONTAL);
  sizerBottom->Add(sizerButtonsHorizontal, wxALL|wxCENTER, 2);
  sizerButtonsHorizontal->Add(buttonUndo, 0, wxALL|wxCENTER, 2);
  sizerButtonsHorizontal->Add(buttonRedo, 0, wxALL|wxCENTER, 2);

  // Status Bar
  CreateStatusBar();
  SetStatusText(wxEmptyString);

  // Setup
  panelTop->Show(false);
  line->Show(false);
  textBoxStartWord->SetFocus();
  sizerAll->Fit(this);
  this->SetSize(400, -1);

  // Threads
  workerQueue = new Queue(this);
  findPathQueue = new Queue(this);
  workerThread = new WorkerThread(workerQueue, 1);
  workerThread->Run();
  findPathThread = new FindPathThread(findPathQueue);
  findPathThread->Run();
}

NeighborsFrame::~NeighborsFrame() {
  delete matchObject;
  delete wordManager;
  delete workerQueue;
  delete findPathQueue;
  delete workerThread;
  delete findPathThread;
}

void NeighborsFrame::findPath() {
  labelFindPathResult->SetLabel(wxEmptyString);
  labelFindPathResultDescription->SetLabel(_("Shortest Path"));
  disableTopWidgets();
  SetStatusText(_("calculating shortest path..."));
  wxString wxStartWord = textBoxStartWord->GetValue();
  wxString wxEndWord = textBoxEndWord->GetValue();
  wxString messageString = wxStartWord + _(" ") + wxEndWord;
  string startWord = WordListUtil::toString(wxStartWord);
  string endWord = WordListUtil::toString(wxEndWord);
  string validateResult = matchObject->validateInput(startWord, endWord);
  if (validateResult == "") {
    findPathQueue->AddJob(ThreadJob(ThreadJob::ID_THREAD_PATH_FROM_START_END_WORDS, messageString), Queue::NORMAL);
  } else {
    wxMessageBox(WordListUtil::toWxString(validateResult), _("Problem"));
    enableTopWidgets();
  }
}

void NeighborsFrame::setComboBox(const wxString& word) {
  int numWords = wordManager->getCurrentIndex() - 1;
  wxString wxStringNumWords = wxString::Format(_("%d"), numWords);
  updateStatusBar(numWords);
  if (combo->GetValue() == labelEndWord->GetLabel()) {
    labelWordListStatus->SetLabel(_("Congratulations!"));
    {
      wxString wxStringShortestPath = wxString::Format(_("%d"), shortestPath);
      wxMessageBox(_("Congratulations! You found a match in " + wxStringNumWords + _(" steps.\n") +
          _("The shortest possible path has ") + wxStringShortestPath + _(" steps.")), _("Winner!"));
    }
    combo->Disable();
    buttonUndo->Disable();
    buttonRedo->Disable();
  } else {
    labelWordListStatus->SetLabel(_("Select Next Word"));
  }
  set<string> nextWordSet = WordListUtil::nextWords(WordListUtil::toString(word), matchObject->getDictionary());
  combo->Clear();
  BOOST_FOREACH(string s, nextWordSet) {
    combo->Append(WordListUtil::toWxString(s));
  }
}

void NeighborsFrame::initializeGame() {
  labelStartWord->SetLabel(wxEmptyString);
  labelEndWord->SetLabel(wxEmptyString);
  labelWordList->SetLabel(wxEmptyString);
  labelWordListStatus->SetLabel(wxEmptyString);
  disableBottomWidgets();
  SetStatusText(_("generating words..."));
  wordManager->reset();
  boost::shared_ptr<StartEndPair> sep(getNextStartEndPair());
  if (!sep) {
    waitingForPair = true;
    return;
  }
  waitingForPair = false;
  startWord = sep->getStartWord();
  endWord = sep->getEndWord();
  shortestPath = sep->getShortestPath();
  labelStartWord->SetLabel(WordListUtil::toWxString(startWord));
  labelEndWord->SetLabel(WordListUtil::toWxString(endWord));
  wordManager->addWord(WordListUtil::toWxString(startWord));
  combo->Clear();
  combo->Append(WordListUtil::toWxString(startWord));
  combo->SetValue(WordListUtil::toWxString(startWord));
  updateWordListLabel(wordManager->getWordList(), wordManager->getCurrentIndex());
  setComboBox(WordListUtil::toWxString(startWord));
  enableBottomWidgets();
}

void NeighborsFrame::updateWordListLabel(const vector<wxString>& v, unsigned int currentIndex) {
  wxString wordList = wxEmptyString;
  for (unsigned int i = 0; i < currentIndex; ++i) {
    wxString word = v[i];
    wordList += (word + _(" "));
  }
  labelWordList->SetLabel(wordList);
  labelWordList->Wrap(labelWordList->GetParent()->GetRect().width - 5);
}

boost::shared_ptr<StartEndPair> NeighborsFrame::getNextStartEndPair() {
  if (startEndPairQueue.empty())
    return boost::shared_ptr<StartEndPair>();
  boost::shared_ptr<StartEndPair> sep(startEndPairQueue.front());
  startEndPairQueue.pop();
  if (startEndPairQueue.size() < START_END_PAIR_CACHE_SIZE)
    workerQueue->AddJob(ThreadJob(ThreadJob::ID_THREAD_REQUEST_MORE, wxEmptyString), Queue::NORMAL);
  return sep;
}

void NeighborsFrame::disableTopWidgets() {
  buttonFindPath->Disable();
  textBoxStartWord->Disable();
  textBoxEndWord->Disable();
}

void NeighborsFrame::disableBottomWidgets() {
  GetToolBar()->EnableTool(ID_NewGame, false);
  combo->Disable();
  buttonUndo->Disable();
  buttonRedo->Disable();
  menuBar->Enable(ID_NewGame, false);
}

void NeighborsFrame::enableTopWidgets() {
  buttonFindPath->Enable();
  buttonFindPath->SetFocus();
  textBoxStartWord->Enable();
  textBoxEndWord->Enable();
}

void NeighborsFrame::enableBottomWidgets() {
  GetToolBar()->EnableTool(ID_NewGame, true);
  combo->Enable();
  buttonUndo->Enable();
  buttonRedo->Enable();
  menuBar->Enable(ID_NewGame, true);
}

void NeighborsFrame::updateStatusBar(int numWords) {
  wxString wxStringNumWords = wxString::Format(_("%d"), numWords);
  if (numWords < 0)
    SetStatusText(wxEmptyString);
  else if (numWords == 1)
    SetStatusText(wxStringNumWords + _(" Word Used"));
  else
    SetStatusText(wxStringNumWords + _(" Words Used"));
}

void NeighborsFrame::OnNewGame(wxCommandEvent& WXUNUSED(event)) {
  initializeGame();
}

void NeighborsFrame::OnPathCalculator(wxCommandEvent& WXUNUSED(event)) {
  if (panelTop->IsShown()) {
    GetToolBar()->ToggleTool(ID_PathCalculator, false);
    menuFile->FindChildItem(ID_PathCalculator)->SetText(_("Show &Path Calculator"));
  } else {
    GetToolBar()->ToggleTool(ID_PathCalculator, true);
    menuFile->FindChildItem(ID_PathCalculator)->SetText(_("Hide &Path Calculator"));
    textBoxStartWord->SetFocus();
  }
  int originalWidth = this->GetSize().GetWidth();
  panelTop->Show(!panelTop->IsShown());
  line->Show(!line->IsShown());
  sizerAll->Fit(this);
  this->SetSize(originalWidth, -1);
}

void NeighborsFrame::OnAbout(wxCommandEvent& WXUNUSED(event)) {
  wxDialog dialog(this, wxID_ANY, wxString(_("About")));
  HtmlWindow* html = new HtmlWindow(&dialog, wxID_ANY, wxDefaultPosition, wxSize(380, 120), wxHW_SCROLLBAR_NEVER);
  html->SetBorders(0);
  html->LoadPage(_("about.htm"));
  html->SetSize(html->GetInternalRepresentation()->GetWidth(), html->GetInternalRepresentation()->GetHeight());
  wxButton* button = new wxButton(&dialog, wxID_OK, _("OK"));
  button->SetDefault();

  // Layout
  wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);
  dialog.SetSizer(sizer);
  sizer->Add(html, 1, wxALL, 10);
  sizer->Add(new wxStaticLine(&dialog, wxID_ANY), 0, wxEXPAND|wxLEFT|wxRIGHT, 10);
  sizer->Add(button, 0, wxALL|wxALIGN_RIGHT, 15);

  // Setup
  sizer->Fit(&dialog);
  dialog.ShowModal();
}

void NeighborsFrame::OnSimpleHelp(wxCommandEvent& WXUNUSED(event)) {
  wxDialog dialog(this, wxID_ANY, wxString(_("Help")));
  HtmlWindow* html = new HtmlWindow(&dialog, wxID_ANY, wxDefaultPosition, wxSize(380, 200), wxHW_SCROLLBAR_NEVER);
  html->SetBorders(0);
  html->LoadPage(_("help.htm"));
  html->SetSize(html->GetInternalRepresentation()->GetWidth(), html->GetInternalRepresentation()->GetHeight());
  wxButton* button = new wxButton(&dialog, wxID_OK, _("OK"));
  button->SetDefault();

  // Layout
  wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);
  dialog.SetSizer(sizer);
  sizer->Add(html, 1, wxALL, 10);
  sizer->Add(new wxStaticLine(&dialog, wxID_ANY), 0, wxEXPAND|wxLEFT|wxRIGHT, 10);
  sizer->Add(button, 0, wxALL|wxALIGN_RIGHT, 15);

  // Setup
  sizer->Fit(&dialog);
  dialog.ShowModal();
}

void NeighborsFrame::OnQuit(wxCommandEvent& WXUNUSED(event)) {
  Close(TRUE);
}

void NeighborsFrame::OnButton(wxCommandEvent& WXUNUSED(event)) {
  findPath();
}

void NeighborsFrame::OnButtonUndo(wxCommandEvent& WXUNUSED(event)) {
  wordManager->undo();
  combo->SetValue(wordManager->getCurrentWord());
  updateWordListLabel(wordManager->getWordList(), wordManager->getCurrentIndex());
  setComboBox(wordManager->getCurrentWord());
}

void NeighborsFrame::OnButtonRedo(wxCommandEvent& WXUNUSED(event)) {
  wordManager->redo();
  combo->SetValue(wordManager->getCurrentWord());
  updateWordListLabel(wordManager->getWordList(), wordManager->getCurrentIndex());
  setComboBox(wordManager->getCurrentWord());
}

void NeighborsFrame::OnComboBox(wxCommandEvent& WXUNUSED(event)) {
  wxString word = combo->GetValue();
  wordManager->addWord(word);
  updateWordListLabel(wordManager->getWordList(), wordManager->getCurrentIndex());
  setComboBox(word);
}

void NeighborsFrame::OnKeyPress(wxKeyEvent& event) {
  if (event.GetKeyCode() == WXK_RETURN) {
    buttonFindPath->SetFocus();
    findPath();
  }
}

void NeighborsFrame::OnThread(wxCommandEvent& event) {
  // From Find Path Thread
  if (event.GetId() == ThreadJob::ID_THREAD_PATH_FROM_START_END_WORDS) {
    wxString steps = wxString::Format(_("%d"), event.GetInt() - 1);
    labelFindPathResult->SetLabel(event.GetString());
    labelFindPathResult->Wrap(labelFindPathResult->GetParent()->GetRect().width - 5);
    if (event.GetInt() == 0) {
      wxMessageBox(_("There is no possible path between these words."), _("Result"));
      labelFindPathResultDescription->SetLabel(_("Shortest Path"));
    } else {
      labelFindPathResultDescription->SetLabel(_("Shortest Path (" + steps + _(" steps)")));
    }
    enableTopWidgets();
    int numWords = wordManager->getCurrentIndex() - 1;
    updateStatusBar(numWords);
    return;
  }

  // From Worker Thread
  int shortestPathFromThread = event.GetInt() - 1;
  wxString w = event.GetString();
  string s = WordListUtil::toString(w);
  string startWordFromThread = WordListUtil::firstWord(s);
  string endWordFromThread = WordListUtil::secondWord(s);
  startEndPairQueue.push(boost::shared_ptr<StartEndPair>(new StartEndPair(startWordFromThread, endWordFromThread, shortestPathFromThread)));
  cout << "(Items in queue: " << startEndPairQueue.size() << ") " << startWordFromThread << " to " << endWordFromThread << " in " << shortestPathFromThread << " steps" << endl;
  if (startEndPairQueue.size() > START_END_PAIR_CACHE_SIZE)
    workerQueue->AddJob(ThreadJob(ThreadJob::ID_THREAD_HAVE_ENOUGH, wxEmptyString), Queue::NORMAL);
  if (waitingForPair)
    initializeGame();
}
