#ifndef NEIGHBORSFRAME_H_
#define NEIGHBORSFRAME_H_

#include <string>
#include <vector>
#include <queue>
#include <set>
#include <wx/wx.h>
#include <wx/statline.h>
#include <wx/event.h>
#include <wx/thread.h>
#include <wx/wxhtml.h>
#include <boost/shared_ptr.hpp>
#include "Match.h"
#include "WordListUtil.h"
#include "WordManager.h"
#include "StartEndPair.h"
#include "WorkerThread.h"
#include "FindPathThread.h"
#include "Queue.h"
#include "HtmlWindow.h"

class NeighborsFrame : public wxFrame {
  public:

    enum {
      ID_NewGame = 1,
      ID_PathCalculator,
      ID_About,
      ID_Help,
      ID_Quit,
      ID_Button,
      ID_ButtonUndo,
      ID_ButtonRedo,
      ID_ComboBox,
      ID_StartThread
    };

    NeighborsFrame(const wxString& title, const wxPoint& pos, const wxSize& size);
    ~NeighborsFrame();
    void findPath();
    void setComboBox(const wxString&);
    void initializeGame();
    void updateWordListLabel(const std::vector<wxString>&, unsigned int);
    boost::shared_ptr<StartEndPair> getNextStartEndPair();
    void disableTopWidgets();
    void disableBottomWidgets();
    void enableTopWidgets();
    void enableBottomWidgets();
    void updateStatusBar(int);
    void OnNewGame(wxCommandEvent&);
    void OnPathCalculator(wxCommandEvent&);
    void OnAbout(wxCommandEvent&);
    void OnSimpleHelp(wxCommandEvent&);
    void OnQuit(wxCommandEvent&);
    void OnButton(wxCommandEvent&);
    void OnButtonUndo(wxCommandEvent&);
    void OnButtonRedo(wxCommandEvent&);
    void OnComboBox(wxCommandEvent&);
    void OnKeyPress(wxKeyEvent&);
    void OnThread(wxCommandEvent&);

  private:
    static const unsigned int START_END_PAIR_CACHE_SIZE = 30;
    int shortestPath;
    bool waitingForPair;
    std::string startWord;
    std::string endWord;
    WordManager* wordManager;
    wxMenu *menuFile;
    wxMenu *menuHelp;
    wxMenuBar *menuBar;
    wxPanel* panel;
    wxPanel* panelTop;
    wxPanel* panelBottom;
    wxStaticText* labelStartWordDescriptionTop;
    wxStaticText* labelEndWordDescriptionTop;
    wxStaticText* labelFindPathResultDescription;
    wxStaticText* labelFindPathResult;
    wxStaticText* labelWordList;
    wxStaticText* labelWordListDescription;
    wxStaticText* labelWordListStatus;
    wxStaticText* labelStartWordDescriptionBottom;
    wxStaticText* labelEndWordDescriptionBottom;
    wxStaticText* labelStartWord;
    wxStaticText* labelEndWord;
    wxStaticLine* line;
    wxTextCtrl* textBoxStartWord;
    wxTextCtrl* textBoxEndWord;
    wxComboBox* combo;
    wxButton* buttonFindPath;
    wxButton* buttonUndo;
    wxButton* buttonRedo;
    Match* matchObject;
    wxBoxSizer* sizerAll;
    wxBoxSizer* sizerTop;
    wxBoxSizer* sizerBottom;
    wxBoxSizer* sizerWordsHorizontal;
    wxBoxSizer* sizerStartVertical;
    wxBoxSizer* sizerEndVertical;
    wxBoxSizer* sizerButtonsHorizontal;
    std::queue<boost::shared_ptr<StartEndPair> > startEndPairQueue;
    Queue* workerQueue;
    Queue* findPathQueue;
    WorkerThread* workerThread;
    FindPathThread* findPathThread;

  DECLARE_EVENT_TABLE()
};

#endif
