#include "Queue.h"

Queue::Queue(wxEvtHandler* parent) :
    parent(parent) {
}

void Queue::AddJob(const ThreadJob& job, const Priority& priority) {
  wxMutexLocker lock(mutexQueue);
  jobs.insert(std::pair<Priority, ThreadJob>(priority, job));
  queueCount.Post();
}

ThreadJob Queue::Pop() {
  ThreadJob element;
  queueCount.Wait();
  mutexQueue.Lock();
  element = (jobs.begin())->second;
  jobs.erase(jobs.begin());
  mutexQueue.Unlock();
  return element;
}

bool Queue::Empty() {
  wxMutexLocker lock(mutexQueue);
  return jobs.empty();
}

void Queue::Report(const ThreadJob::ThreadCommand& command, const wxString& sArg, int iArg) {
  wxCommandEvent event(wxEVT_THREAD, command);
  event.SetString(sArg);
  event.SetInt(iArg);
  parent->AddPendingEvent(event);
}

size_t Queue::Stacksize() {
  wxMutexLocker lock(mutexQueue);
  return jobs.size();
}
