#include "WordSearch.h"

using namespace std;
typedef boost::unordered_map<string, vector<string> >::iterator mapIter;

WordSearch::WordSearch(const string& startWord, const boost::unordered_set<string>& dictionary) :
    startWord(startWord), dictionary(dictionary), numberWordsEverUsed(1) {
  vector<string> v;
  v.push_back(startWord);
  wordsCurrent.insert(pair<string, vector<string> > (startWord, v));
  wordsEverUsed.insert(startWord);
}

void WordSearch::processWord(const string& word, mapIter it) {
  if (dictionary.find(word) != dictionary.end()) {
    pair<boost::unordered_set<string>::iterator, bool> returnedPair;
    returnedPair = wordsEverUsed.insert(word);
    if (returnedPair.second) {
      vector<string> v(it->second);
      v.push_back(word);
      wordsCurrent.insert(pair<string, vector<string> > (word, v));
    }
  }
}

void WordSearch::deleteLetter(mapIter it) {
  for (size_t i = 0; i < it->first.length(); ++i) {
    string temp = it->first;
    processWord(temp.erase(i, 1), it);
  }
}

void WordSearch::changeLetter(mapIter it) {
  for (size_t i = 0; i < it->first.length(); ++i) {
    for (char j = 'a'; j <= 'z'; ++j) {
      string temp = it->first;
      temp[i] = j;
      processWord(temp, it);
    }
  }
}

void WordSearch::addLetter(mapIter it) {
  for (size_t i = 0; i < it->first.length(); ++i) {
    for (char j = 'a'; j <= 'z'; ++j) {
      string temp = it->first;
      // insert letter before existing letter
      processWord(temp.insert(i, 1, j), it);
    }
  }

  for (char j = 'a'; j <= 'z'; ++j) {
    string temp = it->first;
    // add letter after last letter
    processWord(temp.append(1, j), it);
  }
}

void WordSearch::allPermutations() {
  wordsLast = wordsCurrent;
  wordsCurrent.clear();

  // try all possible new words for each word currently in our list
  for (mapIter it = wordsLast.begin(); it != wordsLast.end(); ++it) {
    deleteLetter(it);
    changeLetter(it);
    addLetter(it);
  }
  numberWordsEverUsed = static_cast<int> (wordsEverUsed.size());
}
