#include "WorkerThread.h"

using namespace std;

WorkerThread::WorkerThread(Queue* jobQueue, int id) :
    jobQueue(jobQueue), id(id), matchObject(Match::instance()), requestMore(true) {
  assert(jobQueue);
  wxThread::Create();
}

wxThread::ExitCode WorkerThread::Entry() {
  ThreadJob::ThreadCommand iErr;
  try {
    while (true) // this is the main loop: process jobs until a job handler throws
      OnJob();
  } catch (ThreadJob::ThreadCommand& i) {
    jobQueue->Report(iErr = i, wxEmptyString, id);
  }
  return (wxThread::ExitCode) iErr; // and return exit code
}

void WorkerThread::OnJob() {
  if (!jobQueue->Empty()) {
    ThreadJob job = jobQueue->Pop(); // This will block if the queue is empty
    switch (job.command) {
      case ThreadJob::ID_THREAD_REQUEST_MORE:
        requestMore = true;
        break;
      case ThreadJob::ID_THREAD_HAVE_ENOUGH:
        requestMore = false;
        break;
      default:
        break;
    }
  }
  if (requestMore) {
    string startWord = "";
    string endWord = "";
    int shortestPath = 0;
    do {
      startWord = matchObject->getRandomWordFromDictionary();
      endWord = startWord;
      while (startWord == endWord) {
        endWord = matchObject->getRandomWordFromDictionary();
      }
    } while ((shortestPath = matchObject->shortestPathFromStartWordToEndWord(startWord, endWord).size()) == 0);

    // notify the main thread
    string startEndString = startWord + " " + endWord;
    wxString wxStartEndString(startEndString.c_str(), wxConvUTF8);
    jobQueue->Report(ThreadJob::ID_THREAD_PAIR_COMPLETE, wxStartEndString, shortestPath);
  } else {
    Sleep(1000);
  }
}
